/*
	
	ArpDebug.cpp
	
	Copyright (c)1998 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.
*/

//#if defined(ArpDEBUG)
#define USE_STREAMS 1
//#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include <ArpKernel/ArpDebug.h>
#endif

// ----------------------- DEBUGGING IMPLEMENTATION -----------------------

#if defined(ArpDEBUG)

#include <map>

ostream& ArpDBStream(void)
{
	return cerr;
}

// This is the object that maps module names to their current
// debugging level.  We don't make a static instance of the object,
// because that would introduce dependencies in the link order of
// files.  By doing it this way, the object is created as soon as
// any instrumented code is executed.
static map<ArpString,int>* modlevel = NULL;
static int corelevel = 0;
static bool terminated = false;

class ArpDebugCleanup {
public:
	~ArpDebugCleanup() {
		ArpDL("core", 1, cdb << ADH << "Cleaning up debug statics." << endl);
		delete modlevel;
		modlevel = 0;
		terminated = true;
	}
};

static ArpDebugCleanup debugcleanup;

static inline map<ArpString,int>* level_map(void)
{
	if( terminated ) return 0;
	if( !modlevel ) {
		modlevel = new map<ArpString,int>();
		#if 0
		ArpString one("One"), two("Two");
		cdb << "strcmp(one,two) == " << strcmp(one, two) << endl;
		cdb << "one.Compare(two) == " << one.Compare(two) << endl;
		cdb << "one < two == " << (one < two ? "true" : "false") << endl;
		#endif
	}
	return modlevel;
}

const map<ArpString,int>& ArpDebugLevel(void)
{
	map<ArpString,int>* level = level_map();
	ArpASSERT(level != NULL);
	(*level)["core"] = corelevel;
	return *level;
}

static ArpString ArpDebugName(const char* mod)
{
	ArpASSERT(mod != NULL);
	if( mod == NULL ) return ArpString("");
	
	ArpString retName(mod);
	int length = retName.Length()+1;
	char* name = retName.LockBuffer(length);
	if( name ) {
		char* start = strrchr(name,'/');
		if( !start ) start = name;
		else start++;
		
		char* end = strrchr(name,'.');
		
		length = 0;
		while( *start && (end == 0 || start < end) ) {
			*name = *start;
			name++;
			start++;
			length++;
		}
		
		*name = 0;
		length++;
	}
	
	retName.UnlockBuffer(length);
	return retName;
}

ArpString ArpDebugHeader(const char* mod, int line)
{
	ArpString out("<");
	if( mod ) {
		const char* s1 = strrchr(mod,'/');
		if( s1 != 0 ) mod = s1+1;
		out += mod;
	}
	out += "#";
	out += int32(line);
	out += "> ";
	return out;
}

int ArpDebugLevel(const char* module)
{
	ArpASSERT(module != NULL);
	if( module == 0 ) return 0;

	if( strcmp(module,"core") == 0 ) return corelevel;
	map<ArpString,int>* level = level_map();
	if( level == 0 ) return 0;
	
	const ArpString modName(ArpDebugName(module));
	if( corelevel > 4 ) {
		cdb << ADH << "Name to get module '" << module
			<< "' is '" << modName << "'"
			<< endl;
	}
	
	map<ArpString, int>::iterator item = level->find(modName);
	if( item == level->end() ) {
		if( corelevel >= 1 ) {
			cdb << ADH << "First access of module " << modName
				<< "; setting to level 0" << endl;
		}
		(*level)[modName] = 0;
	}
	
	return (*level)[modName];
}

void ArpDebugLevel(const char* module, int level)
{
	ArpVALIDATE(module != NULL, return);
	ArpVALIDATE(level >= 0, level = 0);

	if( strcmp(module,"core") == 0 ) {
		if( corelevel >= 1 ) {
			cdb << ADH << "Setting module " << module
				<< " to level " << level << endl;
		}
		corelevel = level;
		return;
	}
	map<ArpString,int>* levelmap = level_map();
	if( levelmap == 0 ) {
		cdb << ADH << "ArpDebugLevel: Unable to retrieve level map!" << endl;
		return;
	}
	
	const ArpString modName(ArpDebugName(module));
	if( corelevel > 4 ) {
		cdb << ADH << "Name to set module '" << module
			  << "' is '" << modName << "'"
			  << endl;
	}
	
	if( corelevel >= 1 ) {
		cdb << ADH << "Setting module " << modName
			<< " to level " << level << endl;
	}
	
	(*levelmap)[modName] = level;
}

void ArpDebugDeclare(const char* module, int level)
{
	//ArpDebugLevel(module, level);
#if 1
	ArpVALIDATE(module != NULL, return);
	ArpVALIDATE(level >= 0, level = 0);

	if( strcmp(module,"core") == 0 ) {
		if( corelevel >= 1 ) {
			cdb << ADH << "Initializing module " << module
				<< " to level " << level << endl;
		}
		return;
	}
	map<ArpString,int>* levelmap = level_map();
	if( levelmap == 0 ) {
		cdb << ADH << "ArpDebugDeclare: Unable to retrieve level map!" << endl;
		return;
	}
	
	const ArpString modName(ArpDebugName(module));
	if( corelevel > 4 ) {
		cdb << ADH << "Name to declare module '" << module
			  << "' is '" << modName << "'"
			  << endl;
	}
	
	map<ArpString, int>::iterator item = levelmap->find(modName);
	if( item == levelmap->end() ) {
		if( corelevel >= 1 ) {
			cdb << ADH << "Initializing module " << modName
				<< " to level " << level << endl;
		}
		(*levelmap)[modName] = level;
	} else {
		if( corelevel >= 1 ) {
			cdb << ADH << "Can't initialize module " << modName
				<< " to level " << level << ": already at level "
				<< (*levelmap)[modName] << endl;
		}
	}
#endif
}

void ArpAllDebugLevels(int level)
{
	map<ArpString,int>* levelmap = level_map();
	if( levelmap == 0 ) return;
	
	map<ArpString,int>::iterator i;
	for( i=levelmap->begin(); i!=levelmap->end(); i++ ) {
		ArpDebugLevel(i->first, level);
	}
	ArpDebugLevel("core", level);
}

void ArpParseDBOpts(int32& argc, char** argv)
{
	ArpDL("core", 1, cdb << "Parsing " << argc << " debug arguments..." << endl);
	
	if( argv == 0 ) return;
	
	int i=1, j=1;
	while( i < argc ) {
		ArpDL("core", 1, cdb << "Parse: i=" << i << ", j=" << j
							<< ", argc=" << argc << ", argv[i]="
							<< (argv[i] ? argv[i] : "<NULL>") << endl);
		if( argv[i] && strcmp(argv[i],"--debug") == 0 ) {
			ArpDL("core", 1, cdb << "Found option: " << argv[i] << endl);
			ArpAllDebugLevels(1);
			i++;
		} else if( argv[i] && strncmp(argv[i],"--debug=",8) == 0 ) {
			ArpDL("core", 1, cdb << "Parsing option: " << argv[i] << endl);
			char* str = argv[i] + 8;
			const char* module = 0;
			while( *str ) {
				ArpDL("core", 1, cdb << "Next parse: " << str << endl);
				const char* base = str;
				while( *str >= '0' && *str <= '9' ) str++;
				if( *str == 0 || *str == ':' || *str == ',' ) {
					if( *str != 0 ) {
						*str = 0;
						str++;
					}
					if( module ) {
						ArpDL("core", 1, cdb << "Setting module " << module
											<< " to level " << base << endl);
						ArpDebugLevel(module, atoi(base));
					} else {
						ArpDL("core", 1, cdb << "Setting everything"
											<< " to level " << base << endl);
						ArpAllDebugLevels(atoi(base));
					}
					module = 0;
				} else {
					module = base;
					while( *str != 0 && *str != ':' && *str != ',' ) str++;
					char term = *str;
					if( term ) {
						*str = 0;
						str++;
					}
					ArpDL("core", 1, cdb << "Found module: " << module << endl);
					if( term != ':' ) {
						ArpDL("core", 1, cdb << "Defaulting module to level 1"
											<< endl);
						ArpDebugLevel(module, 1);
						module = 0;
					}
				}
			}
			i++;
		} else if( argv[i] && strcmp(argv[i],"--list-modules") == 0 ) {
			ArpDL("core", 1, cdb << "Found option: " << argv[i] << endl);
			map<ArpString,int>* levelmap = level_map();
			if( levelmap != 0 ) {
				map<ArpString,int>::iterator i;
				for( i=levelmap->begin(); i!=levelmap->end(); i++ ) {
					cdb << "Module '" << i->first
						<< "' is at debug level " << i->second << endl;
				}
				cdb << "Module 'core' is at debug level " << corelevel << endl;
			}
			i++;
		} else if( argv[i] &&
					( strcmp(argv[i],"--help") == 0 ||
					  strcmp(argv[i],"-h") == 0 ||
					  strcmp(argv[i],"-?") == 0 ) ) {
			cdb << "Debug Options: --debug=[module][:level][...] --list-modules"
				<< endl;
			i++;
			j++;
		} else {
			ArpDL("core", 1, cdb << "Skipping argument: "
								<< (argv[i] ? argv[i]:"<NULL>") << endl);
			if( j < i ) argv[j] = argv[i];
			i++;
			j++;
		}
	}
	
	argc = j;
}

#endif

// ----------------------- IOSTREAM FUNCTIONS -----------------------

#ifndef _FONT_H
#include <Font.h>
#endif

#ifndef _ENTRY_H
#include <Entry.h>
#endif

#ifndef _PATH_H
#include <Path.h>
#endif

#ifndef _TEXTVIEW_H
#include <TextView.h>
#endif

#ifndef __BSTRING__
#include <be/support/String.h>
#endif

#include <iostream.h>

ostream& operator << (ostream& os, const BPoint & bp)
{
	os << "BPoint(" << bp.x << "," << bp.y << ")"; return os;
}

ostream& operator << (ostream& os, const BRect & br)
{
	os << "BRect(" << br.left << "," << br.top
     << ")-(" << br.right << "," << br.bottom << ")";
	return os;
}

ostream& operator << (ostream& os, const BFont & fn)
{
	ios::fmtflags fl = os.flags();
	font_family family;
	font_style style;
	fn.GetFamilyAndStyle(&family,&style);
	os << "BFont(id=" << hex << fn.FamilyAndStyle() << dec
		<< ",fam=" << &family[0] << ",sty=" << &style[0]
		<< ",sz=" << fn.Size() << ",shr=" << fn.Shear()
		<< ",rot=" << fn.Rotation() << ",enc=" << fn.Encoding()
		<< ",spc=" << fn.Spacing() << ",fc=" << fn.Face()
		<< ",flg=" << fn.Flags() << ",dir=" << (int)fn.Direction()
		<< ")";
	os.flags(fl);
	return os;
}

ostream& operator << (ostream& os, const BMessenger & o)
{
	ios::fmtflags fl = os.flags();
	BHandler* hnd = NULL;
	BLooper* loop = NULL;
	hnd = o.Target(&loop);
	os << "BMessenger(valid=" << o.IsValid()
		<< hex << ",team=" << o.Team()
		<< ",hnd=" << hnd << ",lp=" << loop
		<< dec << ",local=" << o.IsTargetLocal() << ")";
	os.flags(fl);
	return os;
}
		
ostream& operator << (ostream& os, const BPath & o)
{ os << "BPath(" << (o.Path() ? o.Path() : "<null>") << ")"; return os; }

ostream& operator << (ostream& os, const entry_ref & o)
{
	ios::fmtflags fl = os.flags();
	os << "entry_ref(dev=" << o.device
		<< ",dir=" << hex << (int32)(o.directory>>32)
		<< (int32)(o.directory) << dec
		<< ",name=" << o.name << ")";
	os.flags(fl);
	return os;
}

ostream& operator << (ostream& os, const rgb_color & o)
{
	os << "rgb_color(r=" << (int)o.red << ",g=" << (int)o.green
		<< ",b=" << (int)o.blue << ")";
	return os;
}

ostream& operator << (ostream& os, const pattern & o)
{ os << "pattern(0x" << (void*)&o << ")"; return os; }

ostream& operator << (ostream& os, const BMessage & msg)
{
	return ArpToStream(os, msg);
}

ostream& ArpToStream(ostream& os, const BMessage & msg, const char* prefix)
{
	ios::fmtflags fl = os.flags();
	os << "BMessage: " << (char)((msg.what>>24)&0xFF)
		<< (char)((msg.what>>16)&0xFF)
		<< (char)((msg.what>>8)&0xFF) << (char)(msg.what&0xFF);
	os << " {" << endl << prefix;
	char* name;
	type_code type;
	long count;
	for( int32 i=0; !msg.GetInfo(B_ANY_TYPE,i,&name,&type,&count);
			i++ ) {
		for( int32 j=0; j<count; j++ ) {
			os << i << "." << j << ": " << name << "=";
			switch( type ) {
			case B_CHAR_TYPE: {
				const char* getit=NULL;
				ssize_t rsize=0;
				os << "(CHAR) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_INT8_TYPE: {
				const int8* getit=NULL;
				ssize_t rsize=0;
				os << "(INT8) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_INT16_TYPE: {
				const int16* getit=NULL;
				ssize_t rsize=0;
				os << "(INT16) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_INT32_TYPE: {
				const int32* getit=NULL;
				ssize_t rsize=0;
				os << "(INT32) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_INT64_TYPE: {
				const int64* getit=NULL;
				ssize_t rsize=0;
				os << "(INT64) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (double)(*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_UINT8_TYPE: {
				const uint8* getit=NULL;
				ssize_t rsize=0;
				os << "(UINT8) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_UINT16_TYPE: {
				const uint16* getit=NULL;
				ssize_t rsize=0;
				os << "(UINT16) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_UINT32_TYPE: {
				const uint32* getit=NULL;
				ssize_t rsize=0;
				os << "(UINT32) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_UINT64_TYPE: {
				const uint64* getit=NULL;
				ssize_t rsize=0;
				os << "(UINT64) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (double)(*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_FLOAT_TYPE:
				os << "(FLOAT) " << msg.FindFloat(name,j) << endl << prefix;
				break;
			case B_DOUBLE_TYPE:
				os << "(DOUBLE) " << msg.FindDouble(name,j) << endl << prefix;
				break;
			case B_BOOL_TYPE:
				os << "(BOOL) "
					<< (msg.FindBool(name,j) ? "true" : "false")
					<< endl << prefix;
				break;
			case B_OFF_T_TYPE: {
				const off_t* getit=NULL;
				ssize_t rsize=0;
				os << "(OFF_T) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (double)(*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_SIZE_T_TYPE: {
				const size_t* getit=NULL;
				ssize_t rsize=0;
				os << "(SIZE_T) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_SSIZE_T_TYPE: {
				const ssize_t* getit=NULL;
				ssize_t rsize=0;
				os << "(SSIZE_T) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_POINTER_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(POINTER) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)getit << " rtti="
					/*<< class_name(*getit)*/ << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_OBJECT_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(OBJECT) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)getit << " rtti="
					/*<< class_name(*getit)*/ << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_MESSAGE_TYPE: {
				BMessage getit;
				os << "(BMESSAGE) ";
				if( !msg.FindMessage(name,j,&getit) ) {
					BString newPrefix(prefix);
					newPrefix += "  ";
					ArpToStream(os, getit, newPrefix.String());
					os << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_MESSENGER_TYPE: {
				const BMessenger* getit=NULL;
				ssize_t rsize=0;
				os << "(BMESSENGER) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_POINT_TYPE:
				os << "(POINT) " << msg.FindPoint(name,j)
					<< endl << prefix;
				break;
			case B_RECT_TYPE:
				os << "(RECT) " << msg.FindRect(name,j)
					<< endl << prefix;
				break;
			case B_REF_TYPE: {
				entry_ref getit;
				os << "(ENTRY_REF) ";
				if( !msg.FindRef(name,j,&getit) ) {
					os << (getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_RGB_COLOR_TYPE: {
				const rgb_color* getit=NULL;
				ssize_t rsize=0;
				os << "(RGB_COLOR) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_PATTERN_TYPE: {
				const pattern* getit=NULL;
				ssize_t rsize=0;
				os << "(PATTERN) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << (*getit) << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_ASCII_TYPE: {
				const char* getit=NULL;
				ssize_t rsize=0;
				os << "(ASCII) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					os << getit << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_STRING_TYPE:
				os << "(STRING) " << msg.FindString(name,j)
					<< endl << prefix;
				break;
			case B_MONOCHROME_1_BIT_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(MONOCHROME_1_BIT) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)(getit) << ", size=" << rsize << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_GRAYSCALE_8_BIT_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(GRAYSCALE_8_BIT) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)(getit) << ", size=" << rsize << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_COLOR_8_BIT_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(COLOR_8_BIT) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)(getit) << ", size=" << rsize << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_RGB_32_BIT_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(RGB_32_BIT) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)(getit) << ", size=" << rsize << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_TIME_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(TIME) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)(getit) << ", size=" << rsize << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_RAW_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(RAW) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)(getit) << ", size=" << rsize << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_MIME_TYPE: {
				const char* getit=NULL;
				ssize_t rsize=0;
				os << "(MIME) ";
				if( !msg.FindData(name,type,j,(const void**)&getit,&rsize) ) {
					if( strncmp(name,"text/",5) == 0 ) {
						os << getit << endl << prefix;
					} else if (strcmp(name,"application/x-vnd.Be-text_run_array") == 0 ) {
						int32 mysize = rsize;
						text_run_array* array
							= BTextView::UnflattenRunArray((const void*)getit,&mysize);
						if( array ) {
							os << "count = " << array->count << " {"
								<< endl << prefix;
							for( i=0; i<array->count; i++ ) {
								os << "\toffset="
									<< array->runs[i].offset
									<< " color="
									<< array->runs[i].color
									<< " font="
									<< array->runs[i].font
									<< endl << prefix;
							}
							free(array);
							os << "\t}" << endl << prefix;
						} else {
							os << "can't unflatten / ptr=" << (void*)getit << endl << prefix;
						}
					} else {
						os << "<can't display>" << endl << prefix;
					}
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			case B_ANY_TYPE: {
				const void* getit=NULL;
				ssize_t rsize=0;
				os << "(RAW) ";
				if( !msg.FindData(name,type,j,&getit,&rsize) ) {
					os << (void*)(getit) << ", size=" << rsize << endl << prefix;
				} else {
					os << "<not found>" << endl << prefix;
				}
			} break;
			default: {
				BPath getit;
				if( !msg.FindFlat(name,j,&getit) ) {
					os << "(BPATH) " << (getit) << endl << prefix;
				} else {
					os << "(unknown) #" << type << endl << prefix;
				}
			}
			}
		}
	}
	os << "}";
	os.flags(fl);
	return os;
}
